#!/usr/bin/env python
# File:                Ampel-ZTF/ampel/ztf/t3/complement/ZTFCutoutImages.py
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                18.09.2020
# Last Modified Date:  18.09.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from base64 import b64decode
from collections.abc import Iterable
from typing import Literal

import backoff
import requests
from requests_toolbelt.sessions import (  # type: ignore[import-untyped]
    BaseUrlSession,
)

from ampel.abstract.AbsBufferComplement import AbsBufferComplement
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.struct.T3Store import T3Store


class LSSTCutoutImages(AbsBufferComplement):
    """
    Add cutout images from LSST archive database
    """

    #: Which detection to retrieve cutouts for
    eligible: Literal["first", "last", "brightest", "all"] = "last"

    archive_url: str = (
        "https://ampel-dev.ia.zeuthen.desy.de/api/lsst/archive/v1/"
    )
    insecure: bool = False

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.session = BaseUrlSession(base_url=self.archive_url)

    @backoff.on_exception(
        backoff.expo,
        requests.ConnectionError,
        max_tries=5,
        factor=10,
    )
    @backoff.on_exception(
        backoff.expo,
        requests.HTTPError,
        giveup=lambda e: not isinstance(e, requests.HTTPError)
        or e.response is None
        or e.response.status_code not in {502, 503, 504, 429, 408},
        max_time=60,
    )
    def get_cutout(self, diaSourceId: int) -> None | dict[str, bytes]:
        response = self.session.get(
            f"alert/{diaSourceId}/cutouts", verify=not self.insecure
        )
        if response.status_code == 404:
            return None

        response.raise_for_status()
        json = response.json()
        return {
            k: b64decode(json[k])
            for k in ["cutoutScience", "cutoutTemplate", "cutoutDifference"]
        }

    def complement(self, records: Iterable[AmpelBuffer], t3s: T3Store) -> None:  # noqa: ARG002
        for record in records:
            if (photopoints := record.get("t0")) is None:
                raise ValueError(f"{type(self).__name__} requires t0 records")
            pps = sorted(
                [pp for pp in photopoints if "LSST_DP" in pp.get("tag", [])],
                key=lambda pp: pp["body"]["midpointMjdTai"],
            )
            if not pps:
                return

            if self.eligible == "last":
                candids = [pps[-1]["id"]]
            elif self.eligible == "first":
                candids = [pps[0]["id"]]
            elif self.eligible == "brightest":
                candids = [
                    min(pps, key=lambda pp: pp["body"]["psfFlux"])["id"]
                ]
            else:  # all
                candids = [pp["id"] for pp in pps]
            cutouts = {candid: self.get_cutout(candid) for candid in candids}

            if "extra" not in record or record["extra"] is None:
                record["extra"] = {self.__class__.__name__: cutouts}
            else:
                record["extra"][self.__class__.__name__] = cutouts



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/aux/LSSTFPFilter.py
# License           : BSD-3-Clause
# Author            : Marcus Fenner <mf@physik.hu-berlin.de>
# Date              : 15.09.2021
# Last Modified Date: 15.09.2021
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>


from ampel.abstract.AbsApplicable import AbsApplicable
from ampel.content.DataPoint import DataPoint


class LSSTFPFilter(AbsApplicable):
    """
    Only get LSST's forced photometry datapoints
    """

    def apply(self, arg: list[DataPoint]) -> list[DataPoint]:
        return [el for el in arg if "LSST_FP" in el["tag"]]



import bson

from ampel.queue.AbsProducer import AbsProducer

from .KafkaProducerBase import KafkaProducerBase


class KafkaProducer(KafkaProducerBase[AbsProducer.Item], AbsProducer):
    def serialize(self, item: AbsProducer.Item) -> bytes:
        return bson.encode(
            {"stock": item.stock, "t0": item.t0, "t1": item.t1, "t2": item.t2}
        )



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/ingest/LSSTDataPointShaper.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 20.04.2021
# Last Modified Date: 21.03.2022
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

from collections.abc import Iterable
from typing import Any

from bson import encode

from ampel.abstract.AbsT0Unit import AbsT0Unit
from ampel.content.DataPoint import DataPoint
from ampel.types import StockId
from ampel.util.hash import hash_payload

from .FieldSelection import FieldSelection, StringFilter


class LSSTDataPointShaper(AbsT0Unit):
    """
    This class 'shapes' datapoints in a format suitable
    to be saved into the ampel database
    """

    digest_size: int = 8  # Byte width of datapoint ids

    fields: FieldSelection = FieldSelection.default()
    # Mandatory implementation

    def process(  # type: ignore[override]
        self, arg: Iterable[dict[str, Any]], stock: StockId
    ) -> list[DataPoint]:
        """
        :param arg: sequence of unshaped dps
        """

        ret_list: list[DataPoint] = []
        # Record forced photometry for comp.
        sourceid_list: set[int | str] = set()
        setitem = dict.__setitem__

        for photo_dict in arg:
            tags = ["LSST"]
            if "band" in photo_dict:
                setitem(photo_dict, "band", photo_dict["band"].lower())
                tags.append("LSST_" + photo_dict["band"].upper())
            """
            Non detection limit don't have an identifier.
            """

            selection: None | StringFilter = None

            if "diaSourceId" in photo_dict:
                tags.append("LSST_DP")
                sourceid_list.add(photo_dict["diaSourceId"])
                selection = self.fields.diaSource
            elif "diaForcedSourceId" in photo_dict:
                tags.append("LSST_FP")
                selection = self.fields.diaForcedSource
            elif "diaObjectId" in photo_dict:  # DiaObject
                # diaObjectId is also used in (prv)diaSource and diaForcedPhotometry
                # if other fields are added, check if they contain diaObjectId
                tags.append("LSST_OBJ")
                selection = self.fields.diaObject
            else:
                # Nondetection Limit
                tags.append("LSST_ND")
                selection = self.fields.diaNondetectionLimit

            if selection:
                body = {k: photo_dict[k] for k in selection.filter(photo_dict)}
            else:
                body = photo_dict

            id = hash_payload(
                encode(dict(sorted(body.items()))),
                size=-self.digest_size * 8,
            )
            ret_list.append(
                {"id": id, "stock": stock, "tag": tags, "body": body}  # type: ignore[typeddict-item]
            )

        # Current alert format allows for the same dp to be provided as
        # both source and forcedsource. If so, we here choose FP.
        # (flux values should consistently be the same)
        return [
            dp
            for dp in ret_list
            if not (
                dp["body"].get("diaForcedSourceId") in sourceid_list
                and "LSST_FP" in dp["tag"]
            )
        ]



from typing import Any

from pydantic import FilePath

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.secret.NamedSecret import NamedSecret


class TLSAuthentication(AmpelBaseModel):
    ca: FilePath
    certificate: FilePath
    key: FilePath
    password: NamedSecret[str]

    def librdkafka_config(self) -> dict[str, Any]:
        return {
            "security.protocol": "ssl",
            "ssl.ca.location": str(self.ca),
            "ssl.certificate.location": str(self.certificate),
            "ssl.key.location": str(self.key),
            "ssl.key.password": self.password.get(),
        }



#!/usr/bin/env python
# File:                Ampel-LSST/ampel/lsst/t2/T2GetDiaObject.py
# License:             BSD-3-Clause
# Author:              Marcus Fennner <mf@physik.hu-berlinn.de>
# Date:                22.03.2022
# Last Modified Date:  22.06.2022
# Last Modified By:    jno <jnordin@physik.hu-berlinn.de>

from typing import TYPE_CHECKING, ClassVar

from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit
from ampel.content.DataPoint import DataPoint
from ampel.model.DPSelection import DPSelection
from ampel.protocol.LoggerProtocol import LoggerProtocol
from ampel.types import UBson


class T2GetDiaObject(AbsPointT2Unit):
    """
    Get information from a LSST transient's diaObject to use in other T2
    """

    eligible: ClassVar[DPSelection] = DPSelection(
        filter="LSSTObjFilter", sort="diaObjectId", select="last"
    )
    params: list[str]

    if TYPE_CHECKING:
        logger: LoggerProtocol

    def process(self, datapoint: DataPoint) -> UBson:
        return {param: datapoint["body"].get(param) for param in self.params}



from collections.abc import Iterable

import bson
import confluent_kafka

from ampel.queue.AbsConsumer import AbsConsumer
from ampel.t2.T2QueueWorker import QueueItem

from .KafkaConsumerBase import KafkaConsumerBase


class KafkaConsumer(KafkaConsumerBase, AbsConsumer[QueueItem]):
    def consume(self) -> None | QueueItem:
        message = self._poll(self.stop)
        if message is None:
            return None
        item: QueueItem = bson.decode(message.value())  # type: ignore[assignment]
        item["_meta"] = confluent_kafka.TopicPartition(  # type: ignore[typeddict-unknown-key]
            message.topic(), message.partition(), message.offset() + 1
        )
        return item

    def acknowledge(self, messages: Iterable[QueueItem]) -> None:
        """
        Store offsets of fully-processed messages
        """
        offsets: dict[tuple[str, int], confluent_kafka.TopicPartition] = dict()
        for message in messages:
            meta: confluent_kafka.TopicPartition = message["_meta"]  # type: ignore[typeddict-item]
            key = (meta.topic, meta.partition)
            if key not in offsets or meta.offset > offsets[key].offset:
                offsets[key] = meta
        try:
            self._consumer.store_offsets(offsets=list(offsets.values()))
        except confluent_kafka.KafkaException as exc:
            # librdkafka will refuse to store offsets on a partition that is not
            # currently assigned. this can happen if the group is rebalanced
            # while a batch of messages is in flight. see also:
            # https://github.com/confluentinc/confluent-kafka-dotnet/issues/1861
            err = exc.args[0]
            if err.code() != confluent_kafka.KafkaError._STATE:  # noqa: SLF001
                raise



#!/usr/bin/env python
# File:                Ampel-LSST/ampel/lsst/t2/T2GetAlertId.py
# License:             BSD-3-Clause
# Author:              Marcus Fennner <mf@physik.hu-berlinn.de>
# Date:                31.03.2022
# Last Modified Date:  31.03.2022
# Last Modified By:    Marcus Fennner <mf@physik.hu-berlinn.de>

from collections.abc import Sequence
from typing import ClassVar, Literal

from ampel.abstract.AbsTiedPointT2Unit import AbsTiedPointT2Unit
from ampel.content.DataPoint import DataPoint
from ampel.model.DPSelection import DPSelection
from ampel.model.UnitModel import UnitModel
from ampel.types import UBson
from ampel.view.T2DocView import T2DocView


class T2GetAlertId(AbsTiedPointT2Unit):
    """
    Get first alertId associated with the latest datapoint in a T1
    """

    eligible: ClassVar[DPSelection] = DPSelection(
        filter="LSSTDPFilter", sort="diaSourceId", select="last"
    )
    t2_dependency: Sequence[UnitModel[Literal["T2GetAlertJournal"]]]

    def process(
        self, datapoint: DataPoint, t2_views: Sequence[T2DocView]
    ) -> UBson:
        sourceid = datapoint["body"]["diaSourceId"]
        t0journals: list[dict] = []
        for t2_view in t2_views:
            if t2_view.unit != "T2GetAlertJournal":
                continue
            payload = t2_view.get_payload()
            assert isinstance(payload, list)
            for journal in payload:
                if "upsert" in journal and sourceid in journal["upsert"]:
                    t0journals += [journal]
        if not t0journals:
            return {}
        t0journals.sort(key=lambda x: x["ts"])
        runids = set([t0journal["run"] for t0journal in t0journals])
        if len(runids) > 1:
            self.logger.warn(
                f"Multiple runids found {runids}, assuming latest"
            )
            t0journals = [
                journal
                for journal in t0journals
                if journal["run"] == sorted(runids)[-1]
            ]
        first = t0journals[0]
        return {"ts": first["ts"], "alertId": first["alert"]}



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/alert/load/MultiAvroAlertLoader.py
# License           : BSD-3-Clause
# Author            : Marcus Fenner <mf@physik.hu-berlin.de>
# Date              : 07.05.2021
# Last Modified Date: 14.09.2021
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

from collections.abc import Iterable
from io import BytesIO, IOBase
from typing import no_type_check

from fastavro import reader

from ampel.abstract.AbsAlertLoader import AbsAlertLoader
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.model.UnitModel import UnitModel


class MultiAvroAlertLoader(AbsAlertLoader[BytesIO]):
    """
    Load avro alerts from another AlertLoader.
    This is needed if there are multiple Alerts in a single avro file.
    """

    loader: UnitModel

    def __init__(self, **kwargs) -> None:
        if "loader" in kwargs and isinstance(kwargs["loader"], str):
            kwargs["loader"] = {"unit": kwargs["loader"]}
        super().__init__(**kwargs)

        self.set_alert_source(self.loader)

    def set_alert_source(self, loader) -> None:
        self.alert_loader: AbsAlertLoader[Iterable[IOBase]] = (
            AuxUnitRegister.new_unit(  # type: ignore
                model=loader
            )
        )
        self.next_file()

    @no_type_check
    def next_file(self) -> None:
        self.reader = reader(next(self.alert_loader))

    def __iter__(self):
        return self

    @no_type_check
    def __next__(self) -> BytesIO:
        try:
            return next(self.reader)
        except StopIteration:
            self.next_file()
            return next(self.reader)



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/ingest/LSSTCompilerOptions.py
# License           : BSD-3-Clause
# Author            : Marcus Fenner <mf@physik.hu-berlin.de>
# Date              : 25.08.2021
# Last Modified Date: 28.02.2022
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

from typing import Any

from ampel.model.ingest.CompilerOptions import CompilerOptions


class LSSTCompilerOptions(CompilerOptions):
    stock: dict[str, Any] = {"tag": "LSST"}
    t0: dict[str, Any] = {"tag": "LSST"}
    t1: dict[str, Any] = {"tag": "LSST"}
    state_t2: dict[str, Any] = {"tag": "LSST"}
    point_t2: dict[str, Any] = {"tag": "LSST"}
    stock_t2: dict[str, Any] = {"tag": "LSST"}



from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroDeserializer

from ampel.base.AmpelBaseModel import AmpelBaseModel

from .HttpSchemaRepository import DEFAULT_SCHEMA
from .PlainAvroDeserializer import Deserializer, PlainAvroDeserializer


class SchemaRegistryURL(AmpelBaseModel):
    registry: str

    def deserializer(self) -> Deserializer:
        return AvroDeserializer(SchemaRegistryClient({"url": self.registry}))


class StaticSchemaURL(AmpelBaseModel):
    root_url: str = DEFAULT_SCHEMA

    def deserializer(self) -> Deserializer:
        return PlainAvroDeserializer(self.root_url)


AvroSchema = SchemaRegistryURL | StaticSchemaURL



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/view/LSSTT2Tabulator.py
# License           : BSD-3-Clause
# Author            : Marcus Fenner <mf@physik.hu-berlin.de>
# Date              : 25.05.2021
# Last Modified Date: 05.05.2022
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

import sys
from collections.abc import Iterable, Mapping, Sequence
from typing import Any

import numpy as np
from astropy.table import Table

from ampel.abstract.AbsT2Tabulator import AbsT2Tabulator
from ampel.content.DataPoint import DataPoint
from ampel.types import StockId
from ampel.util.collections import ampel_iter

LSST_BANDPASSES = {
    "u": "lsstu",
    "g": "lsstg",
    "r": "lsstr",
    "i": "lssti",
    "z": "lsstz",
    "y": "lssty",
}


class LSSTT2Tabulator(AbsT2Tabulator):
    convert2jd: bool = True
    zp: float
    allow_nan_flux: bool = False
    # tag priority: lower index -> higher priority
    tags: Sequence[str | int] = ["LSST_FP", "LSST_DP"]
    """ """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._tag_priority = {
            tag: index for index, tag in enumerate(self.tags)
        }

    def get_flux_table(
        self,
        dps: Iterable[DataPoint],
    ) -> Table:
        flux, fluxerr, filtername, tai = self.get_values(
            dps,
            ["psfFlux", "psfFluxErr", "band", "midpointMjdTai"],
            self._tag_priority,
        )
        if self.convert2jd:
            tai = self._to_jd(tai)
        filters = list(map(LSST_BANDPASSES.get, filtername))

        table = Table(
            {
                "time": tai,
                "flux": flux,
                "fluxerr": fluxerr,
                "band": filters,
                # ZP for ELAsTiCC, might need to be corrected for LSST!
                "zp": [self.zp] * len(filters),
                "zpsys": ["ab"] * len(filters),
            },
            dtype=("float64", "float64", "float64", "str", "float64", "str"),
        )

        if self.allow_nan_flux:
            return table
        return table[np.isfinite(table["flux"])]

    def get_positions(
        self, dps: Iterable[DataPoint]
    ) -> Sequence[tuple[float, float, float]]:
        return tuple(
            zip(
                self.get_jd(dps),
                *self.get_values(dps, ["ra", "dec"], self._tag_priority),
                strict=False,
            )
        )

    def get_jd(self, dps: Iterable[DataPoint]) -> Sequence[float]:
        return self._to_jd(
            self.get_values(dps, ["midpointMjdTai"], self._tag_priority)[0]
        )

    @staticmethod
    def _to_jd(dates: Sequence[Any]) -> Sequence[Any]:
        return [date + 2400000.5 for date in dates]

    def get_stock_id(self, dps: Iterable[DataPoint]) -> set[StockId]:
        return set(
            stockid
            for el in dps
            if "LSST" in el["tag"]
            for stockid in ampel_iter(el["stock"])
        )

    def get_stock_name(self, dps: Iterable[DataPoint]) -> list[str]:
        return [str(stock) for stock in self.get_stock_id(dps)]

    @staticmethod
    def _select_dps(
        dps: Iterable[DataPoint], tag_priority: Mapping[str | int, int]
    ) -> Iterable[DataPoint]:
        # select one datapoint per visit with highest tag priority
        selected_dps: dict[int, DataPoint] = {}
        for el in dps:
            if (tag_priority.keys() & el["tag"]) and (
                (key := el["body"]["midpointMjdTai"]) not in selected_dps
                or min(tag_priority.get(t, sys.maxsize) for t in el["tag"])
                < min(
                    tag_priority.get(t, sys.maxsize)
                    for t in selected_dps[key]["tag"]
                )
            ):
                selected_dps[key] = el
        return selected_dps.values()

    @staticmethod
    def get_values(
        dps: Iterable[DataPoint],
        params: Sequence[str],
        tag_priority: Mapping[str | int, int],
    ) -> tuple[Sequence[Any], ...]:
        if tup := tuple(
            map(
                list,
                zip(
                    *(
                        [el["body"][param] for param in params]
                        for el in LSSTT2Tabulator._select_dps(
                            dps, tag_priority
                        )
                    ),
                    strict=False,
                ),
            )
        ):
            return tup
        return tuple([[]] * len(params))



#!/usr/bin/env python
# File:                Ampel-ZTF/ampel/lsst/t0/ReallySimpleLSSTFilter.py
# License:             BSD-3-Clause
# Author:              Marcus Fenner <mf@physik.hu-berlin.de>
# Date:                24.03.2022
# Last Modified Date:  24.03.2022
# Last Modified By:    Marcus Fenner <mf@physik.hu-berlin.de>


from ampel.abstract.AbsAlertFilter import AbsAlertFilter
from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol


class ReallySimpleLSSTFilter(AbsAlertFilter):
    """
    ELASTICC specific filter. It selects alerts based on:
    * numper of previous detections
    """

    # History
    min_ndet: int  # number of previous detections
    max_ndet: None | int  # number of previous detections
    min_tspan: float  # minimum duration of alert detection history [days]
    max_tspan: float  # maximum duration of alert detection history [days]

    def post_init(self):
        # feedback
        for k in self.__annotations__:
            self.logger.info(f"Using {k}={getattr(self, k)}")

        # To make this tenable we should create this list dynamically depending on what entries are required
        # by the filter. Now deciding not to include drb in this list, eg.
        self.keys_to_check = ("midpointMjdTai",)

    def _alert_has_keys(self, photop) -> bool:
        """
        check that given photopoint contains all the keys needed to filter
        """
        for el in self.keys_to_check:
            if el not in photop:
                self.logger.info(None, extra={"missing": el})
                return False
            if photop[el] is None:
                self.logger.info(None, extra={"isNone": el})
                return False
        return True

    # Override
    def process(self, alert: AmpelAlertProtocol) -> None | bool:
        """
        Mandatory implementation.
        To exclude the alert, return *None*
        To accept it, either return
        * self.on_match_t2_units
        * or a custom combination of T2 unit names
        """

        pps = [
            el for el in alert.datapoints if el.get("diaSourceId") is not None
        ]
        if len(pps) < self.min_ndet:
            self.logger.info(None, extra={"nDet": len(pps)})
            return None
        if self.max_ndet and len(pps) > self.max_ndet:
            self.logger.info(None, extra={"nDet": len(pps)})
            return None

        # cut on length of detection history
        detections_jds = [el["midpointMjdTai"] for el in pps]
        last_det = max(detections_jds)
        det_tspan = last_det - min(detections_jds)
        if not (self.min_tspan <= det_tspan <= self.max_tspan):
            self.logger.info(None, extra={"tSpan": det_tspan})
            return None

        self.logger.debug("Alert accepted", extra={"midpointMjdTai": last_det})
        return True



from ampel.lsst.ingest.LSSTCompilerOptions import (
    CompilerOptions,
    LSSTCompilerOptions,
)
from ampel.model.UnitModel import UnitModel
from ampel.template.EasyAlertConsumerTemplate import EasyAlertConsumerTemplate


class LSSTAlertConsumerTemplate(EasyAlertConsumerTemplate):
    supplier: str | UnitModel = UnitModel(
        unit="LSSTAlertSupplier", config={"deserialize": None}
    )
    shaper: str | UnitModel = "LSSTDataPointShaper"
    combiner: str | UnitModel = "LSSTT1Combiner"
    compiler_opts: CompilerOptions = LSSTCompilerOptions()
    muxer: None | str | UnitModel = "LSSTMongoMuxer"



#!/usr/bin/env python

import itertools
from collections.abc import Callable, Iterable, Iterator
from threading import Event
from typing import Any

import confluent_kafka
from pydantic import TypeAdapter

from ampel.abstract.AbsAlertLoader import AbsAlertLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.lsst.kafka.AvroSchema import AvroSchema
from ampel.lsst.kafka.KafkaConsumerBase import KafkaConsumerBase

_get_schema: Callable[[Any], AvroSchema] = TypeAdapter(
    AvroSchema
).validate_python


class KafkaAlertLoader(KafkaConsumerBase, AbsAlertLoader[dict]):
    """
    Load alerts from one or more Kafka topics
    """

    #: Message schema (or url pointing to one)
    avro_schema: None | AvroSchema

    def __init__(self, **kwargs):
        if isinstance(kwargs.get("avro_schema"), str):
            kwargs["avro_schema"] = {"root_url": kwargs["avro_schema"]}

        if avro_schema := kwargs.get("avro_schema"):
            kwargs.setdefault("kafka_consumer_properties", {})[
                "value.deserializer"
            ] = _get_schema(avro_schema).deserializer()
        super().__init__(**kwargs)

        self._it: None | Iterator[dict] = None

    def set_logger(self, logger: AmpelLogger) -> None:
        super().set_logger(logger)

    @staticmethod
    def _add_message_metadata(alert: dict, message: confluent_kafka.Message):
        meta = {}
        timestamp_kind, timestamp = message.timestamp()
        meta["timestamp"] = {
            "kind": (
                "create"
                if timestamp_kind == confluent_kafka.TIMESTAMP_CREATE_TIME
                else "append"
                if timestamp_kind == confluent_kafka.TIMESTAMP_LOG_APPEND_TIME
                else "unavailable"
            ),
            "value": timestamp,
        }
        meta["topic"] = message.topic()
        meta["partition"] = message.partition()
        meta["offset"] = message.offset()
        meta["key"] = message.key()

        alert["__kafka"] = meta
        return alert

    def acknowledge(self, alert_dicts: Iterable[dict]) -> None:
        """
        Store offsets of fully-processed messages
        """
        offsets: dict[tuple[str, int], int] = dict()
        for alert in alert_dicts:
            meta = alert["__kafka"]
            key, value = (meta["topic"], meta["partition"]), meta["offset"]
            if key not in offsets or value > offsets[key]:
                offsets[key] = value
        try:
            self._consumer.store_offsets(
                offsets=[
                    confluent_kafka.TopicPartition(
                        topic, partition, offset + 1
                    )
                    for (topic, partition), offset in offsets.items()
                ]
            )
        except confluent_kafka.KafkaException as exc:
            # librdkafka will refuse to store offsets on a partition that is not
            # currently assigned. this can happen if the group is rebalanced
            # while a batch of messages is in flight. see also:
            # https://github.com/confluentinc/confluent-kafka-dotnet/issues/1861
            err = exc.args[0]
            if err.code() != confluent_kafka.KafkaError._STATE:  # noqa: SLF001
                raise

    def _consume(self) -> Iterator[dict]:
        stop = Event()
        while not stop.is_set():
            message = self._poll(stop)
            if message is None:
                return
            else:
                yield self._add_message_metadata(message.value(), message)

    def alerts(self, limit: None | int = None) -> Iterator[dict]:
        """
        Generate alerts until timeout is reached
        :returns: dict instance of the alert content
        :raises StopIteration: when no alerts recieved within timeout
        """

        yield from itertools.islice(self._consume(), limit)

    def __next__(self) -> dict:
        if self._it is None:
            self._it = self.alerts()
        return next(self._it)



from collections.abc import Iterator
from functools import cached_property
from typing import Annotated, Any

from pydantic import Field
from requests_toolbelt.sessions import (  # type: ignore[import-untyped]
    BaseUrlSession,
)

from ampel.abstract.AbsAlertLoader import AbsAlertLoader
from ampel.log.AmpelLogger import AmpelLogger


class LSSTArchiveAlertLoader(AbsAlertLoader[dict]):
    """
    Load alerts from LSST alert archive
    """

    archive_url: str = (
        "https://ampel-dev.ia.zeuthen.desy.de/api/lsst/archive/v1/"
    )
    insecure: Annotated[
        bool, Field(description="Do not verify HTTPS certificates")
    ] = False

    condition: Annotated[
        str, Field(description="SQL condition for alert query")
    ]
    order_by: Annotated[
        None | str, Field(description="Order by clause for alert query")
    ] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._it: None | Iterator[dict] = None

    def set_logger(self, logger: AmpelLogger) -> None:
        super().set_logger(logger)

    @cached_property
    def session(self) -> BaseUrlSession:
        return BaseUrlSession(base_url=self.archive_url)

    def _build_query(
        self, condition: str, limit: None | int = None
    ) -> dict[str, Any]:
        exclude = [
            "cutoutTemplate",
            "cutoutScience",
            "cutoutDifference",
        ]
        return {
            "exclude": exclude,
            "condition": condition,
            "order": self.order_by,
            "limit": limit,
        }

    def alerts(self, limit: None | int = None) -> Iterator[dict]:
        """
        Generate alerts until timeout is reached
        :returns: dict instance of the alert content
        :raises StopIteration: when no alerts recieved within timeout
        """

        response = self.session.post(
            "display/alerts/query",
            json=self._build_query(self.condition, limit=limit),
            verify=not self.insecure,
        )
        try:
            response.raise_for_status()
        except Exception:
            self.logger.error(
                f"Error querying LSST alert archive: {response.json()}"
            )
            raise

        yield from response.json()

    def __next__(self) -> dict:
        if self._it is None:
            self._it = self.alerts()
        return next(self._it)
